/**
 * This program selects instances for the 2024 edition of the PB24 solvers competition.
 *
 * WARNING: this program is not generic enough. This means that it does not handle all possible cases. Besides, it is absolutely not optimal, nor well written.
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <cstdlib>
#include <stdexcept>
#include <boost/algorithm/string.hpp>
using namespace std;

map<string,string> md5; // maps a filename to its md5 hash
multimap<string,string> md52fname; // maps a md5 to a filename
set<string> discarded; // filenames which are discarded because they are duplicates

void readMD5()
{
  ifstream in("md5.list");
  string hash,fname;
  
  while(in.good())
  {
    in >> hash;
    if(in.fail())
      break;
    while(isspace(in.peek()))
      in.get();
    getline(in,fname);

    //cout << hash << " " << fname << endl;
    md5[fname]=hash;
    md52fname.emplace(hash,fname);
  }
}

bool isDup(const string fname)
{
  if(discarded.find(fname)!=discarded.end())
    return true;

  const auto &[b,e]=md52fname.equal_range(md5[fname]);
  if(b==md52fname.end())
    throw runtime_error("missing MD5 for "+fname);

  for(auto it=b;it!=e;++it)
    if(it->second!=fname)
      discarded.insert(it->second);
  
  return false;
}

string firstFile(const string fname)
{
  const auto &[b,e]=md52fname.equal_range(md5[fname]);
  if(b==md52fname.end())
    throw runtime_error("missing MD5 for "+fname);

  for(auto it=b;it!=e;++it)
    if(discarded.find(it->second)==discarded.end())
      return it->second;

  return "--not found--";
}

struct Instance
{
  string name;
  bool selected;

  Instance(const string &name)
  {
    selected=false;
    this->name=name;
  }
};

int main(int argc, char **argv)
{
  if (argc!=2)
  {
    cout << "syntax: selection <file containing the list>\n";
    exit(1);
  }

  ifstream in(argv[1]);
  vector<Instance> list;
  string tmp,mode;
  int nb,total=0;

  srandom(2024);

#define SQL
#ifdef SQL
  ofstream sql("list.sql");
  int idevent=108;

  if(string(argv[1])=="wbo.list")
    idevent=112;
#endif

  readMD5();
    
  while(in.good())
  {
    list.clear();

    if(in.peek()=='#')
    {
      getline(in,tmp);
      cout << tmp << endl;
      continue;
    }

    in >> nb; // number of instances to select;
    in >> mode; // first or rand
    getline(in,tmp);

    if(mode!="first" && mode!="rand")
      throw runtime_error(string("unexpected mode: ")+mode);

    if (!in.good())
      break;

    while(in.good())
    {
      getline(in,tmp);
      boost::trim(tmp);
      if (tmp.length()==0)
	break;

      if(tmp[0]=='#')
	continue;

      if(!isDup(tmp))
	list.push_back(Instance(tmp));
      else
	cout << "#! duplicate ignored: " << tmp << " (first file is "
	     << firstFile(tmp) << ")\n";
    }

    if(nb>=list.size())
    {
      // select every instance
      for(size_t i=0;i<list.size();++i)
	list[i].selected=true;
    }
    else
    {
      if(mode=="rand")
      {
	// random selection
	int i,n=0;
	while(n<nb)
	{
	  i=random()%list.size();
	  
	  if (!list[i].selected)
	  {
	    list[i].selected=true;
	    ++n;
	  }
	}
      }
      else
      {
	// select the first
	for(int i=0;i<nb;++i)
	  list[i].selected=true;
      }
    }

    if(nb>list.size())
      cout << "## WARNING: not enough instances!\n";

    cout << "## " << nb <<  " " << mode << " instances are selected out of " 
	 << list.size() << ":\n";
    for(int i=0;i<list.size();++i)
      if (list[i].selected)
      {
	++total;
	cout << list[i].name << endl;

#ifdef SQL
	int start=0;
	for(int j=0;j<list[i].name.length();++j)
	  if(list[i].name[j]=='@')
	    start=j+1;

	if(!sql.good())
	  throw runtime_error("can't write to sql file");

	sql << "insert into selectedbenchs (idevent,idbench) select "
	    << idevent << ",idbench from benchs where filename='" 
	    << list[i].name.substr(start) 
	    << "' except select idevent,idbench from selectedbenchs;" << endl;
#endif
      }

    cout << "# done (" << total << " instances selected)\n\n";
  }
}
